; ================================================================
; Soundblaster patch for the Ultima 6 MIDI driver
; Last updated on 9-March-2005
;
; Compiling:
; nasmw -f bin u6sb_patch.asm -o u6sb_patch.bin
;
; Compiling with a different base port for the SB card:
; nasmw -f bin codepatch.asm -o codepatch.bin -dsb_base_port=0x220
; ================================================================

CPU 8086

%ifndef  sb_base_port
    %define  sb_base_port  0x220
%endif

; Soundblaster DSP registers
dsp_reset       equ    6 ; Reset
dsp_read        equ  0xA ; Read Data
wb_status       equ  0xC ; Write Buffer Status
data_available  equ  0xE ; Data Available

; Timer registers
chan_0_counter  equ  0x40
control_word    equ  0x43

; The uncompressed u6roland.drv is 1465 bytes long.
; Since the patch is appended to the end of the uncompressed u6roland.drv,
; all addresses in the patch must be offset by 1465 bytes.
org 1465

; The patched u6roland.drv expects these jumps to be near (3 bytes each)
                jmp     near emulate_mpu_401
                jmp     near write_midi_byte

; ---------------
; Write MIDI byte
; Input:
; ah = MIDI byte
; ---------------
write_midi_byte:
                mov     dx,[cs:base_port]
                add     dx,wb_status
check_status_1:
                in      al,dx
                test    al,0x80
                jnz     check_status_1
                mov     al,0x38 ; SB command: write MIDI byte in polling mode
                out     dx,al
check_status_2:
                in      al,dx
                test    al,0x80
                jnz     check_status_2
                mov     al,ah
                out     dx,al ; write MIDI byte
                retn

; --------------------------
; Emulate MPU-401 commands:
; 1) Reset MPU-401 (0xFF)
; 2) Clear play table (0xB9)
; --------------------------
emulate_mpu_401:
                cmp     ah,0xFF ; MPU-401 command: reset
                jnz     check_b9
                jmp     midi_reset
check_b9:
                cmp     ah,0xB9 ; MPU-401 command: clear play table (all notes off)
                jnz     no_emulation
                jmp     all_notes_off
no_emulation:
                retn

; --------------------------
; Wait about 27 milliseconds
; --------------------------
waste_time:
                push    si
                push    di
                push    bx
                mov     si,0 ; timer tick counter, low word
                mov     di,0 ; timer tick counter, high word
                ; control word = 00 00 011 0
                ; counter 0, counter latching, mode 3 (square wave rate generator), binary 16 bit
                mov     al,6
                out     control_word,al
                ; bx := current timer value
                in      al,chan_0_counter ; read low timer byte
                mov     bl,al
                in      al,chan_0_counter ; read high timer byte
                mov     bh,al
read_timer_loop:
                mov     cx,bx
                ; control word = 00 00 011 0
                ; counter 0, counter latching, mode 3 (square wave rate generator), binary 16 bit
                mov     al,6
                out     control_word,al
                ; bx := current timer value
                in      al,chan_0_counter ; read low timer byte
                mov     bl,al
                in      al,chan_0_counter ; read high timer byte
                mov     bh,al
                sub     cx,bx
                jnc     no_timer_wrap
                add     cx,(1193181 / 60) ; highest timer value + 1 (60 Hz)
no_timer_wrap:
                add     si,cx
                jnc     no_counter_wrap
                inc     di
no_counter_wrap:
                cmp     di,0
                jb      read_timer_loop
                cmp     si,0x8000
                jb      read_timer_loop
                pop     bx
                pop     di
                pop     si
                retn

; ---------
; Reset DSP
; ---------
midi_reset:
                ; Write a 1 to the SB reset port (2x6h)
                mov     dx,[cs:base_port]
                add     dx,dsp_reset
                mov     al,1
                out     dx,al

                ; Wait at least 3 microseconds
                call    near waste_time

                ; Write a 0 to the SB reset port (2x6h)
                mov     dx,[cs:base_port]
                add     dx,dsp_reset
                mov     al,0
                out     dx,al

                ; Read from the Data Available port (2xEh) until bit 7 == 1
                mov     dx,[cs:base_port]
                add     dx,data_available
check_data_available:
                in      al,dx
                test    al,0x80
                jz      check_data_available

                ; Wait for 0xAA on the Read Data port (2xAh)
                mov     dx,[cs:base_port]
                add     dx,dsp_read
check_read_data:
                in      al,dx
                cmp     al,0xAA
                jnz     check_read_data

; -------------
; All notes off
; -------------
all_notes_off:
                mov     cx,0 ; current channel
channel_off:
                mov     ah,0xB0 ; channel mode message
                add     ah,cl
                call    near write_midi_byte
                mov     ah,0x7B ; all notes off
                call    near write_midi_byte
                mov     ah,0 ; all notes off
                call    near write_midi_byte
                inc     cx
                cmp     cx,0x10
                jb      channel_off
                retn

; ---------
; Variables
; ---------
base_port  dw  sb_base_port
