/*
 *  ega2bmp.c - converts raw 320x200 EGA images to Windows bmp's
 *
 *  Copyright (C) 2002  Marc Winterrowd
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Conversion utility for Ultima 4 (PC version) image files.
 * How to use it:
 * 1) use u4decode to decompress one of the .ega files (note that not all .ega files are compressed)
 * 2) use this utility to convert the decompressed file to a .bmp
 */

#include <stdio.h>
#include <stdlib.h>

#define width 320
#define height 200
#define palette_entries 256

/* forward function declarations */
int convert_u4ega_to_bmp(char* u4ega_filename, char* bmp_filename);
long get_filesize(FILE *input_file);
void long_to_array4(long p1, unsigned char* p2);
void int_to_array2(int p1, unsigned char* p2);
void copy_bytes(unsigned char *source, unsigned char *dest, long count);
void set_palette_entry(unsigned char *rgb_quad, int index, unsigned char red, unsigned char green, unsigned char blue);

typedef struct _BITMAPFILEHEADER
{
    unsigned char bfType[2]; /* "BM" */
    unsigned char bfSize[4]; /* specifies the size of the file in bytes */
    unsigned char bfReserved1[2]; /* 0 */
    unsigned char bfReserved2[2]; /* 0 */
    unsigned char bfOffBits[4]; /* specifies the offset from the beginning of the file to the bitmap data */
} BITMAPFILEHEADER;

typedef struct _BITMAPINFOHEADER
{
    unsigned char biSize[4];
    unsigned char biWidth[4];
    unsigned char biHeight[4];
    unsigned char biPlanes[2];
    unsigned char biBitCount[2];
    unsigned char biCompression[4];
    unsigned char biSizeImage[4];
    unsigned char biXPelsPerMeter[4];
    unsigned char biYPelsPerMeter[4];
    unsigned char biClrUsed[4];
    unsigned char biClrImportant[4];
} BITMAPINFOHEADER;

typedef struct _BW_RGBQUAD
{
    unsigned char rgbquad[palette_entries*4];
} BW_RGBQUAD;

int main (int argc, char *argv[])
{
    switch (argc)
    {
    case 3:
        return(convert_u4ega_to_bmp(argv[1],argv[2]));
        break;
    default:
        printf("Usage:\n");
        printf("2 arguments - convert decompressed .ega file (arg1) to bmp (arg2).\n");
        return(EXIT_FAILURE);
        break;
    }
}

int convert_u4ega_to_bmp(char* u4ega_filename, char* bmp_filename)
{
    FILE *u4ega_file, *bmp_file;

    if (strcmp (u4ega_filename, bmp_filename) == 0)
    {
        printf("Input and output files must not have identical names.\n");
        return(EXIT_FAILURE);
    }
    else
    {
        if (!(u4ega_file=fopen(u4ega_filename,"rb"))||!(bmp_file=fopen(bmp_filename,"wb")))
        {
            printf("Couldn't open \"%s\" or \"%s\" or both.\n", u4ega_filename, bmp_filename);
            return(EXIT_FAILURE);
        }
        else
        {
            unsigned char *u4ega_mem, *bmp_mem;
            long u4ega_filesize, bmp_filesize;

            /* loop counters */
            int i;
            long il,jl;

            long bits_read,headers;

            BITMAPFILEHEADER bmp_bitmapfileheader;
            BITMAPINFOHEADER bmp_bitmapinfoheader;
            BW_RGBQUAD bmp_bw_rgbquad;
            unsigned char *palpoint;

            /* allocate mem: font and bmp */
            u4ega_filesize = get_filesize(u4ega_file);
            bmp_filesize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + sizeof(BW_RGBQUAD) + width*height;

            u4ega_mem = (unsigned char *) malloc(u4ega_filesize);
            bmp_mem = (unsigned char *) malloc(bmp_filesize);

            /* read font file into mem */
            fread(u4ega_mem, 1, u4ega_filesize, u4ega_file);

            /*
             * create bmp
             */

            /* file header */
            bmp_bitmapfileheader.bfType[0] = 'B';
            bmp_bitmapfileheader.bfType[1] = 'M';
            long_to_array4(bmp_filesize, bmp_bitmapfileheader.bfSize);
            int_to_array2(0,bmp_bitmapfileheader.bfReserved1);
            int_to_array2(0,bmp_bitmapfileheader.bfReserved2);
            long_to_array4(sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+sizeof(BW_RGBQUAD),bmp_bitmapfileheader.bfOffBits);

            /* info header */
            long_to_array4(sizeof(BITMAPINFOHEADER),bmp_bitmapinfoheader.biSize);
            long_to_array4(width,bmp_bitmapinfoheader.biWidth);
            long_to_array4(height,bmp_bitmapinfoheader.biHeight);
            int_to_array2(1,bmp_bitmapinfoheader.biPlanes);
            int_to_array2(8,bmp_bitmapinfoheader.biBitCount);
            long_to_array4(0,bmp_bitmapinfoheader.biCompression);
            long_to_array4(0,bmp_bitmapinfoheader.biSizeImage);
            long_to_array4(0,bmp_bitmapinfoheader.biXPelsPerMeter);
            long_to_array4(0,bmp_bitmapinfoheader.biYPelsPerMeter);
            long_to_array4(0,bmp_bitmapinfoheader.biClrUsed);
            long_to_array4(0,bmp_bitmapinfoheader.biClrImportant);

            /* create palette */
            palpoint = (unsigned char*)&bmp_bw_rgbquad.rgbquad;
            /* the first 16 entries correspond to the EGA palette */
            set_palette_entry(palpoint, 0, 0x00, 0x00, 0x00);
            set_palette_entry(palpoint, 1, 0x00, 0x00, 0x80);
            set_palette_entry(palpoint, 2, 0x00, 0x80, 0x00);
            set_palette_entry(palpoint, 3, 0x00, 0x80, 0x80);
            set_palette_entry(palpoint, 4, 0x80, 0x00, 0x00);
            set_palette_entry(palpoint, 5, 0x80, 0x00, 0x80); /* dark magenta */
            set_palette_entry(palpoint, 6, 0x80, 0x80, 0x00);
            set_palette_entry(palpoint, 7, 0xC0, 0xC0, 0xC0);
            set_palette_entry(palpoint, 8, 0x80, 0x80, 0x80); /* dark gray */
            set_palette_entry(palpoint, 9, 0x00, 0x00, 0xFF);
            set_palette_entry(palpoint, 10, 0x00, 0xFF, 0x00);
            set_palette_entry(palpoint, 11, 0x00, 0xFF, 0xFF); /* cyan */
            set_palette_entry(palpoint, 12, 0xFF, 0x00, 0x00);
            set_palette_entry(palpoint, 13, 0xFF, 0x00, 0xFF);
            set_palette_entry(palpoint, 14, 0xFF, 0xFF, 0x00);
            set_palette_entry(palpoint, 15, 0xFF, 0xFF, 0xFF);

            /* fill the remaining 256-16 entries with 4 zeroes */
            for (i=16; i < palette_entries; i++)
            {
                bmp_bw_rgbquad.rgbquad[i*4+0] = i;
                bmp_bw_rgbquad.rgbquad[i*4+1] = i;
                bmp_bw_rgbquad.rgbquad[i*4+2] = i;
                bmp_bw_rgbquad.rgbquad[i*4+3] = 0x00;
            }

            copy_bytes(bmp_bitmapfileheader.bfType,bmp_mem,sizeof(BITMAPFILEHEADER));
            copy_bytes(bmp_bitmapinfoheader.biSize,bmp_mem+sizeof(BITMAPFILEHEADER),sizeof(BITMAPINFOHEADER));
            copy_bytes(bmp_bw_rgbquad.rgbquad,bmp_mem+sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER),sizeof(BW_RGBQUAD));

            bits_read = 0;
            headers = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + sizeof(BW_RGBQUAD);

            /* convert u4ega into bmp */
            for (il = 0; il < height; il++)
            {
                for (jl = 0; jl < width; jl++)
                {
                    unsigned char pixel;
                    pixel = u4ega_mem[bits_read/8];
                    pixel = (pixel >> (4 - (bits_read % 8))) & 0xf;
                    bmp_mem[headers + (height-1-il)*width + jl] = pixel;
                    bits_read += 4;
                }
            }

            /* write bmp */
            fwrite(bmp_mem, 1, bmp_filesize, bmp_file);

            /* deallocate mem */
            free(u4ega_mem);
            free(bmp_mem);

            /* close files */
            fclose(u4ega_file);
            fclose(bmp_file);
        }
    }

    return(EXIT_SUCCESS);
}

/*
 * Auxiliary functions
 */

long get_filesize(FILE *input_file)
{
    long file_length;
    fseek(input_file, 0, SEEK_END);
    file_length = ftell(input_file);
    fseek(input_file, 0, SEEK_SET);
    return(file_length);
}

void long_to_array4(long p1, unsigned char* p2)
{
    int i;

    for (i=0; i < 4; i++)
    {
        p2[i] = (p1 >> (i*8)) & 0xff;
    }
}

void int_to_array2(int p1, unsigned char* p2)
{
    int i;

    for (i=0; i < 2; i++)
    {
        p2[i] = (p1 >> (i*8)) & 0xff;
    }
}

void copy_bytes(unsigned char *source, unsigned char *dest, long count)
{
    long il;

    for (il=0; il < count; il++)
    {
        dest[il] = source[il];
    }
}

void set_palette_entry(unsigned char *rgb_quad, int index, unsigned char red, unsigned char green, unsigned char blue)
{
    rgb_quad[index*4+0] = blue;
    rgb_quad[index*4+1] = green;
    rgb_quad[index*4+2] = red;
    rgb_quad[index*4+3] = 0;
}
